:root {
  --stale-color: rgb(254,0,82);
  --refresher-color: rgb(0,197,158);
}

.cell.code_cell .out_prompt_overlay::before {
  position: absolute;
  display: block;
  top: 3px;
  left: -6px;
  width: 5px;
  height: calc(100% + 2px);
  content: '';
}

.cell.code_cell .input_prompt::before {
  position: absolute;
  display: block;
  top: -1px;
  left: -1px;
  width: 5px;
  height: calc(100% + 2px);
  content: '';
}

.cell.code_cell.stale-cell .input_prompt::before {
  border: 1px solid var(--stale-color);
}

.cell.code_cell.stale-cell.selected .input_prompt::before, .stale-cell.jupyter-soft-selected .input_prompt::before {
  background-color: var(--stale-color);
}

.cell.code_cell.stale-cell .input_prompt:hover::before {
  background-color: var(--stale-color);
}

.cell.code_cell.linked-stale .input_prompt::before {
  background-color: var(--stale-color);
}

.cell.code_cell.linked-fresh .out_prompt_overlay::before {
  background-color: var(--stale-color);
}

.cell.code_cell.refresher-cell .input_prompt::before {
  border: 1px solid var(--refresher-color);
}

.cell.code_cell.refresher-cell.selected .input_prompt::before, .cell.code_cell.refresher-cell.jupyter-soft-selected .input_prompt::before {
  background-color: var(--refresher-color);
}

.cell.code_cell.refresher-cell .input_prompt:hover::before {
  background-color: var(--refresher-color);
}

.cell.code_cell.linked-refresher .input_prompt::before {
  background-color: var(--refresher-color);
}

/* put input / output cells later since they should take precedence */

.cell.code_cell.fresh-cell .out_prompt_overlay::before {
  border: 1px solid var(--stale-color);
}

.cell.code_cell.fresh-cell.selected .out_prompt_overlay::before, .cell.code_cell.fresh-cell.jupyter-soft-selected .out_prompt_overlay::before {
  background-color: var(--stale-color);
}

.cell.code_cell.fresh-cell .out_prompt_overlay:hover::before {
  background-color: var(--stale-color);
}

.cell.code_cell.refresher-input-cell .input_prompt::before {
  border: 1px solid var(--refresher-color);
}

.cell.code_cell.refresher-input-cell.selected .input_prompt::before, .cell.code_cell.refresher-input-cell.jupyter-soft-selected .input_prompt::before {
  background-color: var(--refresher-color);
}

.cell.code_cell.refresher-input-cell .input_prompt:hover::before {
  background-color: var(--refresher-color);
}
